<?php

namespace Drupal\swapcard_content\Form;

use Symfony\Component\DependencyInjection\ContainerInterface;

use Drupal\Core\Security\TrustedCallbackInterface;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Queue\QueueWorkerManager;

use Drupal\Component\Utility\NestedArray;

use Drupal\swapcard\Form\SwapcardConfigForm;
use Drupal\swapcard\Plugin\SwapcardPluginManager;

/**
 * SwapcardContentConfigForm object definition.
 */
class SwapcardContentConfigForm extends SwapcardConfigForm implements TrustedCallbackInterface {

  /**
   * Drupal\Core\Queue\QueueWorkerManager definition.
   *
   * @var \Drupal\Core\Queue\QueueWorkerManager
   */
  protected $queueManager;

  /**
   * Constructs \Drupal\swapcard_content\Form\SwapcardContentConfigForm object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   Entity type manager interface.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   Cache interface.
   * @param \Drupal\swapcard\Plugin\SwapcardPluginManager $swapcard_plugin_manager
   *   Swapcard plugin manager.
   * @param \Drupal\Core\Queue\QueueWorkerManager $queue_manager
   *   QueueWorkerManager instance.
   */
  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, CacheBackendInterface $cache, SwapcardPluginManager $swapcard_plugin_manager, QueueWorkerManager $queue_manager) {
    parent::__construct($config_factory, $entity_type_manager, $cache, $swapcard_plugin_manager);
    $this->queueManager = $queue_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('cache.default'),
      $container->get('plugin.manager.swapcard'),
      $container->get('plugin.manager.queue_worker')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function trustedCallbacks() {
    return [
      'processEntities',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {

    $form = parent::buildForm($form, $form_state);

    $config = $this->config('swapcard_content.settings');
    $values = $form_state->getValues();

    $form['swapcard_content'] = [
      '#type' => 'details',
      '#title' => $this->t('Swapcard Content settings'),
      '#open' => TRUE,
      '#tree' => TRUE,
    ];

    // Gather the number of events in the form already.
    $num_events = $form_state->get('num_events');
    // We have to ensure that there is at least one widget.
    if ($num_events === NULL) {
      if (is_array($config->get('events')) && count($config->get('events')) > 1) {
        $num_events = count($config->get('events'));
      }
      else {
        $num_events = 1;
      }
    }

    $form_state->set('num_events', $num_events);

    // Container for multiple Events.
    $form['swapcard_content']['events'] = [
      '#type' => 'container',
      '#tree' => TRUE,
      '#attributes' => [
        'id' => 'swapcard-events-wrapper',
      ],
    ];

    for ($delta = 1; $delta <= $num_events; $delta++) {

      $index = $delta - 1;
      $event = NULL;

      if (is_array($config->get('events')) && !empty($config->get('events'))) {
        $event = isset($config->get('events')[$index]) && !empty($config->get('events')[$index]) ? $config->get('events')[$index] : NULL;
      }
      else {
        if (isset($values['events']) && !empty($values['events'])) {
          $event = isset($values['events'][$index]) && !empty($values['events'][$index]) ? $values['events'][$index] : NULL;
        }
      }

      // Custom composite element.
      $form['swapcard_content']['events'][$index] = [
        '#type' => 'textfield',
        '#title' => $this->t('Swapcard Event ID'),
        '#default_value' => $event,
        '#attributes' => [
          'id' => 'swapcard-events-' . $index,
        ],
      ];
    }

    $form['swapcard_content']['events']['add_event'] = [
      '#type' => 'submit',
      '#value' => $this->t('Add Event'),
      '#limit_validation_errors' => [],
      '#submit' => [
        [get_class($this), 'addEventSubmit'],
      ],
      '#weight' => 20,
      '#ajax' => [
        'callback' => [get_class($this), 'ajaxCallback'],
        'wrapper' => 'swapcard-events-wrapper',
      ],
    ];

    if ($num_events > 1) {

      $form['swapcard_content']['events']['remove_event'] = [
        '#type' => 'submit',
        '#value' => $this->t('Remove Event'),
        '#limit_validation_errors' => [],
        '#submit' => [
          [get_class($this), 'removeEventSubmit'],
        ],
        '#weight' => 20,
        '#ajax' => [
          'callback' => [get_class($this), 'ajaxCallback'],
          'wrapper' => 'swapcard-events-wrapper',
        ],
      ];
    }

    $form['swapcard_content']['event_content'] = [
      '#type' => 'details',
      '#title' => $this->t('Event content'),
      '#description' => $this->t("Select event's content entities to import for each event."),
      '#open' => FALSE,
    ];

    $form['swapcard_content']['event_content']['entities'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Event entities'),
      '#default_value' => $config->get('entities'),
      '#description' => $this->t("Note that, as per Swapcard design <em>Speakers</em> are child property of sessions."),
      '#pre_render' => [
        [
          get_class($this), 'processEntities',
        ],
      ],
      '#options' => [
        'swapcard_exhibitor' => $this->t('Exhibitors'),
        'swapcard_session' => $this->t('Sessions'),
        'swapcard_speaker' => $this->t('Speakers'),
      ],
    ];

    $node_types = $config->get('node_types');
    $field_swapcard_session_type = NULL;
    $field_swapcard_location = NULL;

    foreach ($node_types as $fields) {
      if (isset($fields['field_swapcard_session_type']) && isset($fields['field_swapcard_session_type']['id'])) {
        $field_swapcard_session_type = $fields['field_swapcard_session_type']['id'];
      }
      if (isset($fields['field_swapcard_location']) && isset($fields['field_swapcard_location']['id'])) {
        $field_swapcard_location = $fields['field_swapcard_location']['id'];
      }
    }

    $form['swapcard_content']['event_content']['fields'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Event fields'),
      '#description' => $this->t('Handle fields with pre-defined values in Swapcard app, a dropdown type mostly. With specifying its ID found in Swapcard app here, we are able to pull that pre-defined values and set them as "allowed_values" for these fields/dropdowns in Drupal. That way those fields are always synchronised with remote configuraton.'),
      '#open' => FALSE,
    ];

    $form['swapcard_content']['event_content']['fields']['field_swapcard_session_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t("Session's <em>Type</em> field ID"),
      '#default_value' => $field_swapcard_session_type,
    ];

    $form['swapcard_content']['event_content']['fields']['field_swapcard_location'] = [
      '#type' => 'textfield',
      '#title' => $this->t("Session's <em>Location</em> field ID"),
      '#default_value' => $field_swapcard_location,
    ];

    $default_uid = NULL;

    if ($config->get('default_uid')) {
      $default_uid = $config->get('default_uid');
    }
    else {
      $default_uid = $form_state->getValue(['swapcard_content', 'default_uid']) ? $form_state->getValue([
        'swapcard_content',
        'default_uid',
      ]) : NULL;
    }

    $node_author = $default_uid ? $this->entityTypeManager->getStorage('user')->load($default_uid) : NULL;

    $form['swapcard_content']['default_uid'] = [
      '#type' => 'entity_autocomplete',
      '#title' => $this->t('Swapcard nodes author'),
      '#description' => $this->t('Select user which becomes author of Swapcard nodes.'),
      '#target_type' => 'user',
      '#default_value' => $node_author,
      '#required' => TRUE,
    ];

    $form['swapcard_content']['cron'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Sync on cron'),
      '#default_value' => $config->get('cron'),
      '#description' => $this->t('Check this to run synchronisation on each cron.'),
    ];

    $form['swapcard_content']['advanced'] = [
      '#type' => 'details',
      '#title' => $this->t('Advanced'),
      '#description' => $this->t('Some advanced features, mostly for testing and debugging, but also for purging data.'),
      '#open' => FALSE,
    ];

    $form['swapcard_content']['advanced']['purge'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Purge entities'),
      '#default_value' => $config->get('purge'),
      '#description' => $this->t('When checked the obsolete entities, such as Exhibitors, Sessions as well as possibly Speakers, would be deleted. Each of such does not have a relation and/or existence on Swapcard anymore, but previously had, so we would like want to delete it from our image here in Drupal. Think of it as reverse check also may be resources demanding. <strong>Use AT YOUR OWN risk, it is in <em>alpha</em> stage by now</strong>, especially when it is about Speakers.'),
    ];

    // Kill switch, when true does not save anything in Drupal.
    // A big note - when using this QueueWorker still writes data in database
    // and will try to deliver all of it on the next cron.
    $form['swapcard_content']['advanced']['dry_run'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Dry run'),
      '#default_value' => $config->get('dry_run'),
      '#description' => $this->t('If true no content will be saved in Drupal yet all the other operations will run.'),
    ];

    $form['swapcard_content']['advanced']['dump'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Dump API response (YML)'),
      '#default_value' => $config->get('dump'),
      '#description' => $this->t('When checked "response" data which came from API will be saved into YML file in <em>public://swapcard/tests</em> folder. This is disabled when the next option is checked (loading tests).'),
      '#id' => 'swapcard-content-dump',
      '#states' => [
        'enabled' => [
          ':input[id="swapcard-content-test"]' => ['checked' => FALSE],
        ],
      ],
    ];

    $form['swapcard_content']['advanced']['test'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use dumped response (YML)'),
      '#default_value' => $config->get('test'),
      '#description' => $this->t('When checked "response" data will come locally, from YML files previously dumped in <em>public://swapcard/tests</em>. This is disabled when the previous option is checked (dumping tests). This is the way to avoid too many calls to API but also it may come very benefitial when it is about deleting stuff, such as values for fields and/or even some of the main 3 entities from parent event.'),
      '#id' => 'swapcard-content-test',
      '#states' => [
        'disabled' => [
          ':input[id="swapcard-content-dump"]' => ['checked' => TRUE],
        ],
      ],
    ];

    // Synchronise data with remote. Run Queue in a batch.
    $form['actions']['batch_sync'] = [
      '#type' => 'submit',
      '#submit' => ['::batchSyncSubmit'],
      '#value' => $this->t('Sync'),
    ];

    // Purge any obsolete data (if event was removed previously for instance).
    $form['actions']['purge_data'] = [
      '#type' => 'submit',
      '#submit' => ['::purgeDataSubmit'],
      '#value' => $this->t('Purge'),
      '#disabled' => TRUE,
    ];

    // Clear any leftover queue tables.
    $form['actions']['clear'] = [
      '#type' => 'submit',
      '#submit' => ['::clearQueuesSubmit'],
      '#value' => $this->t('Clear'),
    ];

    // Buttons explanation.
    $buttons_markup = $this->t("<p><strong>Save configuration</strong> does default operation but in case when <em>Event fields</em> are set submit will also refresh those fields' values.</p><p><strong>Sync</strong> button will synchronise data in Drupal, mirroring what is currently on Swapcard app for Events defined above, as well as selected Entities for each. </p><p><strong>Purge</strong> is disabled and possibly meant for purging all data related to some event that was just removed, using <em>remove event</em> button above. Attention - this will delete all of the related Exhibitors and Sessions, as well as some Speakers (orphans) that do not have any parent Session anymore. Likewise orphan Media entities (images) will be removed.</p><p><strong>Clear</strong> button will clear any leftover queues in database tables, just these related to this module and starting with <em>swapcard_content_queue_</em>.</p>");

    $form['actions_description'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Actions'),
      '#description' => $buttons_markup,
    ];

    return $form;
  }

  /**
   * Process entities checkboxes.
   *
   * Basically to disable speakers when no sessions selected.
   *
   * @param array $element
   *   Checkboxes form element.
   */
  public static function processEntities(array $element) {
    $element['swapcard_session']['#id'] = 'swapcard-content-session-id';
    $element['swapcard_speaker']['#states'] = [
      'enabled' => [
        ':input[id="swapcard-content-session-id"]' => ['checked' => TRUE],
      ],
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

    $config = $this->configFactory->getEditable('swapcard_content.settings');

    // Take care of checkboxes and user.
    $default_uid = $form_state->getValue(['swapcard_content', 'default_uid']);
    $cron = $form_state->getValue(['swapcard_content', 'cron']);
    $dump = $form_state->getValue(['swapcard_content', 'advanced', 'dump']);
    $test = $form_state->getValue(['swapcard_content', 'advanced', 'test']);
    $dry_run = $form_state->getValue(['swapcard_content', 'advanced', 'dry_run']);
    $purge = $form_state->getValue(['swapcard_content', 'advanced', 'purge']);

    $config->set('default_uid', $default_uid);
    $config->set('cron', $cron);
    $config->set('dump', $dump);
    $config->set('test', $test);
    $config->set('dry_run', $dry_run);
    $config->set('purge', $purge);

    // Take care of event child entities checboxes.
    $event_content = $form_state->getValue([
      'swapcard_content',
      'event_content',
      'entities',
    ]);
    $save_entities = array_filter($event_content, function ($value) {
      return is_string($value) ?? $value;
    });

    $config->set('entities', $save_entities);

    // Take care of event fields entries.
    $event_fields = $form_state->getValue([
      'swapcard_content',
      'event_content',
      'fields',
    ]);
    $field_ids = [];
    $update_config = [];
    $node_types = $config->get('node_types');
    foreach ($node_types as $bundle => $fields) {
      $update_config[$bundle] = $fields;
      foreach ($event_fields as $field_name => $field_value) {
        if (isset($fields[$field_name])) {
          if (!empty($field_value)) {
            $update_config[$bundle][$field_name]['id'] = $field_value;
            $field_ids[$field_value] = $field_value;
          }
        }
      }
    }
    $config->set('node_types', $update_config);

    // Take care of events entries.
    $events = $form_state->getValue(['swapcard_content', 'events']);
    $save_events = [];

    if (!empty($events)) {
      foreach ($events as $delta => $event_id) {
        if (is_numeric($delta) && !empty($event_id)) {
          $save_events[$delta] = trim($event_id);
        }
      }
      // REFRESH defined fields' allowed_values function.
      if (!empty($field_ids) && !empty($save_events)) {
        $this->queueManager->createInstance('swapcard_content_queue_swapcard_fields')->updateFields($save_events, $field_ids, TRUE);
      }
    }

    // Save config and return parent method.
    $config->set('events', $save_events);
    $config->save();
    return parent::submitForm($form, $form_state);
  }

  /**
   * Process bulk synchronisation with data from remote.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function batchSyncSubmit(array &$form, FormStateInterface $form_state) {

    $plugin_manager = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');

    $entities = $plugin_manager->checkEventEntities();

    if (empty($entities)) {
      $plugin_manager->processRequest('swapcard_content_queue_swapcard_event', 'swapcard_event', 'events', TRUE);
    }
    else {
      if (isset($entities['entity_key']) && isset($entities['swapcard_callback'])) {
        $plugin_manager = $this->queueManager->createInstance('swapcard_content_queue_' . $entities['entity_key']);
        $plugin_manager->processRequest('swapcard_content_queue_swapcard_event', $entities['entity_key'], $entities['swapcard_callback'], TRUE);
      }
    }
  }

  /**
   * Remove all queue tables, to start over.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function clearQueuesSubmit(array &$form, FormStateInterface $form_state) {
    $plugin = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $queues = [
      'swapcard_content_queue_swapcard_event',
      'swapcard_content_queue_swapcard_exhibitor',
      'swapcard_content_queue_swapcard_session',
      'swapcard_content_queue_swapcard_speakers',
      'swapcard_content_queue_swapcard_fields',
      'swapcard_content_queue_swapcard_media',
      'swapcard_content_queue_swapcard_delete',
    ];

    foreach ($queues as $queue) {
      $plugin->queueFactory->get($queue)->deleteQueue();
    }
    $plugin->messenger->addStatus($this->t('Queue tables cleared.'));
  }

  /**
   * Purge-obsolete submit.
   *
   * @todo this one.
   *
   * @param array $form
   *   The form where the settings form is being included in.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function purgeDataSubmit(array &$form, FormStateInterface $form_state) {
    $plugin = $this->queueManager->createInstance('swapcard_content_queue_swapcard_event');
    $plugin->messenger->addWarning($this->t('Still TODO!'));
  }

  /**
   * Callback for all ajax actions.
   *
   * Returns parent container element for each group.
   */
  public static function ajaxCallback(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    $parents = array_slice($trigger['#parents'], 0, -1);
    $element = NestedArray::getValue($form, $parents);
    return $element;
  }

  /**
   * Add Event Submit callback.
   */
  public static function addEventSubmit(array &$form, FormStateInterface $form_state) {
    $num_events = $form_state->get('num_events');
    $delta = $num_events + 1;
    $form_state->set('num_events', $delta);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Remove Event Submit callback.
   */
  public static function removeEventSubmit(array &$form, FormStateInterface $form_state) {
    $num_events = $form_state->get('num_events');
    $delta = $num_events - 1;
    $form_state->set('num_events', $delta);
    $form_state->setRebuild(TRUE);
  }

}
