<?php

namespace Drupal\swapcard_content\Plugin\QueueWorker;

use Symfony\Component\Yaml\Yaml;

/**
 * Creates QueueWorker for Swapcard fields.
 *
 * @QueueWorker(
 *   id = "swapcard_content_queue_swapcard_fields",
 *   title = @Translation("Swapcard Fields"),
 *   cron = {"time" = 90}
 * )
 */
class SwapcardQueueWorkerFields extends SwapcardQueueWorkerBase {

  /**
   * {@inheritdoc}
   */
  public function processItem($data) {

    $options = [];
    foreach ($data as $field_id => $field) {
      $options[$field_id] = $this->fetchRemoteFields($field, 'event');
    }
    if (!empty(array_values($options))) {
      $this->messenger->addStatus($this->t('Events fields data successfully synchronised.'));
    }
    return $options;
  }

  /**
   * Update fields that have pre-defined values, fetched from remote.
   *
   * @param array $events
   *   An array with events IDs.
   * @param array $field_ids
   *   An array with unique swapcard field IDs.
   * @param bool $batch
   *   An array with events IDs.
   *
   * @return array
   *   Array with key|Labels for select list field.
   */
  public function updateFields(array $events = [], array $field_ids = [], bool $batch = FALSE) {

    $parent_config = $this->configFactory->get('swapcard.settings');
    $guzzle_options = $parent_config->get('guzzle_options');

    if ($parent_config->get('api_key') && !empty($guzzle_options['base_uri'])) {

      $events = !empty($events) ? $events : $this->config->get('events');

      // Note that Swapcard field's options (key|value)
      // are the same across all events.
      if (!empty($events) && isset($events[0]) && !empty($events[0])) {

        $batch_items = [];
        $output = [];

        if (empty($field_ids)) {
          foreach ($this->config->get('node_types') as $fields) {
            foreach ($fields as $field_name => $field) {
              if (isset($field['id'])) {
                $field_ids[$field_name] = $field['id'];
              }
            }
          }
        }

        foreach ($events as $event_id) {

          $batch_items[$event_id] = [];

          foreach ($field_ids as $drupal_field_name => $field_id) {
            $batch_items[$event_id][$drupal_field_name] = [
              'event_id' => $event_id,
              'field_id' => $field_id,
              'field_type' => 'list_string',
            ];
          }
          if ($batch) {
            $this->queueFactory->get('swapcard_content_queue_swapcard_fields')->createItem($batch_items[$event_id]);
          }
          else {
            foreach ($batch_items[$event_id] as $drupal_field_name => $item) {
              $output[$drupal_field_name] = $this->fetchRemoteFields($item, 'event');
            }
          }
        }
        if (!empty(array_values($batch_items))) {
          if ($batch) {
            $this->queueUi->batch(['swapcard_content_queue_swapcard_fields']);
          }
          else {
            return $output;
          }
        }
      }
      else {
        $this->apiStatus('warning', 'events');
      }
    }
    else {
      $this->apiStatus('error', 'api');
    }

  }

  /**
   * Get Swapcard app fields definitions.
   *
   * @param array $data
   *   Properties for this field.
   * @param string $swapcard_callback
   *   Swapcard API callback function name.
   *
   * @return array
   *   Associative array with key|Label pairs for select list type of field.
   */
  protected function fetchRemoteFields(array $data, string $swapcard_callback) {

    if (!is_array($data) || empty($data)) {
      return [];
    }

    $cid = 'swapcard_content:' . $data['event_id'] . ':' . $data['field_id'];

    if ($cached = $this->cache->get($cid)) {
      if (!empty($cached->data)) {
        return $cached->data;
      }
    }

    $options = [];

    // TEST - add test response array to the queue.
    // Comes from YML file, instad of calling a remote API.
    if ($this->config->get('test') || $this->testPath) {

      $test_file_path = static::DUMP_PATH . '/swapcard_fields.test.yml';
      if (file_exists($test_file_path)) {
        $response['data'][$swapcard_callback] = Yaml::parseFile($test_file_path);
      }
      else {
        $this->messenger->addError($this->t('Specified test file <em>@path</em> does not exist.', [
          '@path' => $test_file_path,
        ]));
        return [];
      }
    }
    else {

      $parent_config = $this->configFactory->get('swapcard.settings');
      $guzzle_options = $parent_config->get('guzzle_options');

      $swapcard_plugin_manager = $this->swapcardPluginManager->createInstance('swapcard_fields', ['all']);
      $args = [
        'id' => '\"' . $data['event_id'] . '\"',
      ];

      $callback_options = [
        'body' => $swapcard_plugin_manager->queryString($swapcard_callback, $args),
      ];

      $response = $swapcard_plugin_manager->post($guzzle_options['base_uri'], $callback_options);
    }

    if (is_array($response) && isset($response['data']) && count($response['data']) > 0 && isset($response['data'][$swapcard_callback])) {

      // DUMP TEST file, if that's set on config.
      if ($this->config->get('dump')) {
        $this->dumpResponse($response['data'][$swapcard_callback], 'swapcard_fields.test.yml');
      }

      foreach ($response['data'] as $response_data) {

        if (isset($response_data['fieldDefinitions']) && !empty($response_data['fieldDefinitions'])) {

          foreach ($response_data['fieldDefinitions'] as $field_definition) {

            if (isset($field_definition['id']) && isset($data['field_id']) && $field_definition['id'] == $data['field_id']) {
              switch ($data['field_type']) {

                case 'list_string':

                  if (isset($field_definition['optionsValues']) && !empty($field_definition['optionsValues'])) {
                    foreach ($field_definition['optionsValues'] as $optionsValue) {
                      if (isset($optionsValue['value']) && !empty($optionsValue['value'])) {
                        $options[$optionsValue['value']] = isset($optionsValue['translations']) && !empty($optionsValue['translations']) && isset($optionsValue['translations'][0]['value']) ? $optionsValue['translations'][0]['value'] : $optionsValue['value'];
                      }
                    }
                  }
                  break;
              }
            }
          }
        }
      }
    }

    if (!empty($options)) {
      // Set this to cache.
      $cache_tags = [$cid];
      $this->cache->set($cid, $options, $this->cache::CACHE_PERMANENT, $cache_tags);
    }
    return $options;
  }

}
